/* Copyright (C) 2000 Stefan Karrmann <S.Karrmann@gmx.net>
 * All Rights reserved.
 */

/* $Id: generate.c,v 1.5 2000/02/26 18:22:12 sk Exp $
 * $Log: generate.c,v $
 * Revision 1.5  2000/02/26 18:22:12  sk
 * Changed for binary secrets
 *
 * Revision 1.4  2000/02/23 21:19:58  sk
 * First working Version
 *
 * Revision 1.2  2000/02/11 22:26:22  sk
 * Kommentarfehler bearbeitet, miese Zufallszahlenschnittstelle
 *
 */

#include "sharesecret.h"
#include <limits.h>

/* eval the polynomial sum(coefficients[i] * x ** i, i=0..(degree - 1))
 * using the horner scheme in the field Z/(modulus Z)
 */

inline
void
eval_polynomial_by_horner (mpz_t val, ulong x, mpz_t modulus,
			   mpz_t * coefficients, ulong degree)
{
  mpz_init (val);
  for (degree--; degree > 0; degree--)
    {
      mpz_add (val, val, coefficients[degree]);
      mpz_mul_ui (val, val, x);
      mpz_mod (val, val, modulus);
    }
  mpz_add (val, val, coefficients[0]);
  mpz_mod (val, val, modulus);
}

/* generates a random polynom p of degree min-1 with p(0) = secret
 * and evaluates it such that result[i] = p(i+1)
 * initializes elements x, y and modulus, callee must (mpz_)clear them.
 */

void
generate (mpz_t ** x, mpz_t ** y, mpz_t modulus, 
	  const mpz_t secret, ulong min, ulong size)
{
  // degree of polynomial: min-1

  mpz_t * coefficients;
  ulong i, j;
  size_t len;

  /* precondition */
  I(min <= size);
  
  /* init */
  * x = xmalloc (sizeof (mpz_t) * size);
  * y = xmalloc (sizeof (mpz_t) * size);
  coefficients = xmalloc (sizeof (mpz_t) * min);
  mpz_init (modulus);

  /* len = 1 + floor ( max (0, log (secret)) / log (2));
     i.e. len = position of MSB
  */
  len =  mpz_sizeinbase (secret, 2);
  if (len > ULONG_MAX - 2)
    error (EXIT_FAILURE, ERR_WARN,
	   _("Secret to long (or too many needed parts)"));

  /* calculate i = ceil (ld (min))
     The multiplicative group of the finite field  Z/pZ (p prime) is
     cyclic. Thus, x ** p - x = 0 in this field, i.e. polynomials are
     at most of degree (p-1). This should be greater than the number
     of needed parts of the secret, e.g. min - 1. The next to lines
     garantees this.
  */
  for ( i = 0, j = min + 1; j > 0; i++, j = j >> 1);
  if (len < i - 2) len = i - 2;
  generate_prime (modulus, len + 2);
  if (HAVE_OPT( VERBOSE))
    fprintf (stderr, "len = %d\n", len);

  /* generate */
  for (i = 1; i < min; i++)	/* use secret as coefficient[0] */
    {
      gen_random (coefficients[i], len);
      if (HAVE_OPT( VERBOSE))
	{
	  fprintf (stderr, "coeff[%d] = ", i);
	  mpz_out_str (stderr, 10, coefficients[i]);
	  fputc ('\n', stderr);
	}
    }
  mpz_init_set (coefficients[0], secret);

  /* evaluate the polynomial size times using the Horner scheme */
  for (i = 0; i < size; i++)
    {
      mpz_init_set_ui ((* x)[i], i + 1);
      eval_polynomial_by_horner ((* y)[i], i + 1, modulus, coefficients, min);
    }

  /*  resulting: done while evaluating the polynomial */
  /* clear */
  for (i = 0; i < min; i++)
    mpz_clear (coefficients[i]);
  free (coefficients);
}

void
xor_generate (mpz_t ** result, const mpz_t secret, size_t size)
{
  ulong i, len;
  mpz_t xors;

  /* init */
  * result = xmalloc (sizeof(mpz_t) * size);
  mpz_init (xors);

  /* len = 1 + floor ( max (0, log (secret)) / log (2));
     i.e. len = position of MSB
  */
  len =  mpz_sizeinbase (secret, 2);

  /* generate */
  for (i = 1; i < size; i++)
    {
      gen_random ((* result)[i], len);
      mpz_xor (xors, xors, (* result)[i]);
    }
  mpz_init ((* result)[0]);
  mpz_xor ((* result)[0], xors, secret);

  /* cleanup */
  mpz_clear (xors);
}
